﻿
function ColorPickerHSL(id, onUpdate, initialColorString) {
    var canvas = document.getElementById(id);
    var ctx = canvas.getContext("2d");

    // conversion functions directly from http://www.w3.org/TR/css3-color/#hsl-color
    function hsl2rgb(h, s, l) {
        var m2 = (l <= 0.5) ? l * (s + 1) : l + s - l * s;
        var m1 = l * 2 - m2;
        var r = hue2rgb(m1, m2, h + 1 / 3);
        var g = hue2rgb(m1, m2, h);
        var b = hue2rgb(m1, m2, h - 1 / 3);
        return { r: r, g: g, b: b };
    }

    function hue2rgb(m1, m2, h) {
        if (h < 0) h = h + 1;
        if (h > 1) h = h - 1;
        if (h * 6 < 1) return m1 + (m2 - m1) * h * 6;
        if (h * 2 < 1) return m2;
        if (h * 3 < 2) return m1 + (m2 - m1) * (2 / 3 - h) * 6;
        return m1;
    }
    
    function rgb2hsl(r, g, b) {
        var maxColor = Math.max(r, g, b);
        var minColor = Math.min(r, g, b);

        var h = 0;
        var s = 0;
        var l = (maxColor + minColor) / 2;
        
        if (maxColor != minColor) {
            if (l < 0.5)
                s = (maxColor - minColor) / (maxColor + minColor);
            else {
                s = (maxColor - minColor) / (2.0 - maxColor - minColor);
            }

            if (r == maxColor)
                h = (g - b) / (maxColor - minColor);
            else if (g == maxColor)
                h = 2.0 + (b - r) / (maxColor - minColor);
            else if (b == maxColor)
                h = 4.0 + (r - g) / (maxColor - minColor);

            if (h < 0)
                h += 6;
            
            h /= 6;
        }

        return ({ h: h, s: s, l: l });
    }

    this.setColorString = function (cssString) {
        SetColorFromString(cssString);
    };

    //  layout values
    var canvasHeight = canvas.height;
    var canvasWidth = canvas.width;
    var margin = 8;
    var lPickerWidth = 18;
    var hslPickerHeight = canvasHeight - margin * 4 - lPickerWidth;   // allow 8 px top and bottom, 16 pixels between, and lPickerWidth for alpha bar
    var hsPickerWidth = canvasWidth - margin * 4 - lPickerWidth;   // allow 8 px left and right, 16 pixels between, and lPickerWidth for lightness bar

    //  default values
    var hue = 1;
    var saturation = 1;
    var lightness = 0.5;
    var alpha = 1;

    function CreateHueSaturationArray() {
        var array = ctx.createImageData(hsPickerWidth, hslPickerHeight);
        var bits = array.data;

        var index = 0;
        for (var y = 0; y < array.height; ++y) {
            for (var x = 0; x < array.width; ++x) {
                var bit = hsl2rgb(x / (array.width - 1), 1 - y / (array.height - 1), 0.5);
                bits[index++] = bit.r * 255;
                bits[index++] = bit.g * 255;
                bits[index++] = bit.b * 255;
                bits[index++] = 255;
            }
        }

        return array;
    }

    function CreateLightnessArray() {
        var array = ctx.createImageData(lPickerWidth, hslPickerHeight);
        var bits = array.data;

        var index = 0;
        for (var y = 0; y < array.height; ++y) {
            var bit = hsl2rgb(hue, saturation, 1 - y / (array.height - 1));

            for (var x = 0; x < array.width; ++x) {
                bits[index++] = bit.r * 255;
                bits[index++] = bit.g * 255;
                bits[index++] = bit.b * 255;
                bits[index++] = 255;
            }
        }

        return array;
    }

    function CreateAlphaArray() {
        var array = ctx.createImageData(hsPickerWidth, lPickerWidth);
        var bits = array.data;

        var index = 0;
        var grayValues = [0.75, 1, 0.75, 1];
        var bit = hsl2rgb(hue, saturation, lightness);
        for (var y = 0; y < array.height; ++y) {
            var yCheck = Math.floor((y % 12) / 6);
            for (var x = 0; x < array.width; ++x) {
                var xCheck = Math.floor((x % 12) / 6);
                var xAlpha = (array.width - x) / (array.width - 1);
                var grayValue = grayValues[yCheck + xCheck] * (1 - xAlpha);
                bits[index++] = (bit.r * xAlpha + grayValue) * 255;
                bits[index++] = (bit.g * xAlpha + grayValue) * 255;
                bits[index++] = (bit.b * xAlpha + grayValue) * 255;
                bits[index++] = 255;
            }
        }

        return array;
    }

    var hsArray = CreateHueSaturationArray();
    var lArray = CreateLightnessArray();
    var aArray = CreateAlphaArray();

    function getHueSaturationXY() {
        return {
            x: 8 + Math.round(hue * (hsArray.width - 1)),
            y: 8 + Math.round((1 - saturation) * (hsArray.height - 1))
        };
    }

    function DrawHueSaturationControl() {
        var hsXY = getHueSaturationXY();

        ctx.clearRect(0, 0, hsArray.width + margin * 2, hsArray.height + margin * 2);

        ctx.putImageData(hsArray, margin, margin);

        ctx.save();
        ctx.translate(0.5, 0.5);

        ctx.strokeStyle = "White";
        ctx.lineWidth = 1;
        ctx.strokeRect(hsXY.x - 7, hsXY.y - 7, 15, 15);
        ctx.fillStyle = "hsl(" + Math.round(hue * 360).toString() + "," + Math.round(saturation * 100).toString() + "%," + Math.round(lightness * 100).toString() + "%)";
        ctx.fillRect(hsXY.x - 6, hsXY.y - 6, 13, 13);
        ctx.strokeStyle = "Black";
        ctx.lineWidth = 1;
        ctx.strokeRect(hsXY.x - 6, hsXY.y - 6, 13, 13);

        ctx.restore();
    };

    function getLightnessXY() {
        return {
            x: canvasWidth - margin - lArray.width,
            y: margin + Math.round((1 - lightness) * (lArray.height - 1))
        };
    }

    function DrawLightnessControl() {
        var lXY = getLightnessXY();

        var lArrayLeft = lXY.x;
        var lArrayRight = lArrayLeft + lArray.width;

        ctx.clearRect(lArrayLeft - margin, 0, margin + lArray.width + margin, margin + lArray.height + margin);

        ctx.putImageData(lArray, lArrayLeft, margin);

        ctx.beginPath();

        ctx.moveTo(lArrayLeft, lXY.y);
        ctx.lineTo(lArrayLeft - 5, lXY.y - 5);
        ctx.lineTo(lArrayLeft - 5, lXY.y + 5);
        ctx.lineTo(lArrayLeft, lXY.y);

        ctx.moveTo(lArrayRight, lXY.y);
        ctx.lineTo(lArrayRight + 5, lXY.y - 5);
        ctx.lineTo(lArrayRight + 5, lXY.y + 5);
        ctx.lineTo(lArrayRight, lXY.y);

        ctx.fillStyle = "#555";
        ctx.fill();
    };

    function DrawAlphaControl() {
        var yTop = canvasHeight - margin - aArray.height;
        var yBottom = yTop + aArray.height;
        var xIndicator = margin + 0.5 + Math.round((1 - alpha) * (aArray.width - 1));

        ctx.clearRect(0, yTop - margin, margin + aArray.width + margin, margin + aArray.height + margin);
        ctx.putImageData(aArray, margin, yTop);

        ctx.beginPath();

        ctx.moveTo(xIndicator, yTop);
        ctx.lineTo(xIndicator - 5, yTop - 5);
        ctx.lineTo(xIndicator + 5, yTop - 5);
        ctx.lineTo(xIndicator, yTop);

        ctx.moveTo(xIndicator, yBottom);
        ctx.lineTo(xIndicator - 5, yBottom + 5);
        ctx.lineTo(xIndicator + 5, yBottom + 5);
        ctx.lineTo(xIndicator, yBottom);

        ctx.fillStyle = "#555";
        ctx.fill();
    }

    function DrawBlackWhiteControl() {
        var topLeftX = canvasWidth - margin - lArray.width;
        var topLeftY = canvasHeight - margin - aArray.height;
        var bottomRightX = canvasWidth - margin;
        var bottomRightY = canvasHeight - margin;

        ctx.save();

        ctx.fillStyle = "white";
        ctx.fillRect(bottomRightX - 12, bottomRightY - 12, 12, 12);
        ctx.strokeStyle = "black";
        ctx.lineWidth = 1;
        ctx.strokeRect(bottomRightX - 12 - 0.5, bottomRightY - 12 - 0.5, 12, 12);

        ctx.fillStyle = "black";
        ctx.fillRect(topLeftX, topLeftY, 12, 12);

        ctx.restore();
    };


    var movingHueSaturation = false;
    var movingLightness = false;
    var movingAlpha = false;

    function ChangeHueSaturationToCanvasCoords(xy) {

        hue = Math.max(0, Math.min(1, (xy.x - margin) / (hsArray.width - 1)));
        saturation = Math.max(0, Math.min(1, 1 - (xy.y - margin) / (hsArray.height - 1)));

        //  if the user is manipulating a color but lightness is at black or white, set it to 50% for them
        if (lightness == 0 || lightness == 1)
            lightness = 0.5;

        //  same kind of helper for 0 alpha
        if (alpha == 0) {
            alpha = 1;
        }

        DrawHueSaturationControl();

        lArray = CreateLightnessArray();    // update lightness array for new hue, saturation
        DrawLightnessControl();

        aArray = CreateAlphaArray();        // update alpha array, too
        DrawAlphaControl();

        UpdateReadout();
    }

    function SetHSLA(h, s, l, a) {
        hue = h;
        saturation = s;
        lightness = l;
        alpha = a;

        DrawHueSaturationControl();

        lArray = CreateLightnessArray();    // update lightness array for new hue, saturation
        DrawLightnessControl();

        aArray = CreateAlphaArray();        // update alpha array, too
        DrawAlphaControl();

        UpdateReadout();
    }

    function SetColorFromString(cssString) {
        var hsl = { h: 0, s: 0, l: 0 }, a = 1;

        var threeDigitHashMatch = cssString.match(/^#([0-9a-f]{1})([0-9a-f]{1})([0-9a-f]{1})$/i);
        if (threeDigitHashMatch) {
            cssString = "#" + threeDigitHashMatch[1] + threeDigitHashMatch[1] + threeDigitHashMatch[2] + threeDigitHashMatch[2] + threeDigitHashMatch[3] + threeDigitHashMatch[3];
        }

        var hashMatch = cssString.match( /^#([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})$/i );
        if (hashMatch) {
            hsl = rgb2hsl(parseInt(hashMatch[1], 16) / 255, parseInt(hashMatch[2], 16) / 255, parseInt(hashMatch[3], 16) / 255);
        }

        var rgbaMatch = cssString.match( /^rgba?\(\s*([\d\.]+),\s*([\d\.]+),\s*([\d\.]+)(?:,\s*([\d\.]+))?\s*\)$/ );
        if (rgbaMatch) {
            hsl = rgb2hsl(parseFloat(rgbaMatch[1]) / 255, parseFloat(rgbaMatch[2]) / 255, parseFloat(rgbaMatch[3]) / 255);
            if (rgbaMatch[4])
                a = parseFloat(rgbaMatch[4]);
        }

        var hslaMatch = cssString.match( /^hsla?\(\s*([\d\.]+),\s*([\d\.]+)%,\s*([\d\.]+)%(?:,\s*([\d\.]+))?\s*\)$/ );
        if (hslaMatch) {
            hsl = { h: parseFloat(hslaMatch[1]) / 360, s: parseFloat(hslaMatch[2]) / 100, l: parseFloat(hslaMatch[3]) / 100 };
            if (hslaMatch[4])
                a = parseFloat(hslaMatch[4]);
        }

        if (cssString.match(/^white$/i))
            hsl = { h: 0, s: 1, l: 1 };
        else if (cssString.match(/^transparent$/i))
            a = 0;

        SetHSLA(hsl.h, hsl.s, hsl.l, a);
    }

    function ChangeLightnessToCanvasCoords(xy) {
        lightness = Math.max(0, Math.min(1, 1 - (xy.y - margin) / (lArray.height - 1)));

        //  if the user is manipulating lightness but alpha is zero, flip alpha to 1
        if (alpha == 0) {
            alpha = 1;
        }

        DrawLightnessControl();

        DrawHueSaturationControl();     // contains a color swatch that depends on lightness

        aArray = CreateAlphaArray();    // update alpha array for change in lightness
        DrawAlphaControl();

        UpdateReadout();
    }

    function ChangeAlphaToCanvasCoords(xy) {
        alpha = Math.max(0, Math.min(1, 1 - (xy.x - margin) / (aArray.width - 1)));

        DrawAlphaControl();

        UpdateReadout();
    }

    /*    function normalizedColorToHexString() {
    var ret = "#";
    for (var i = 0; i < arguments.length; i++) {
    var arg = Math.round(255 * arguments[i]);
    ret += (arg < 16 ? "0" : "") + arg.toString(16);
    }
    return ret.toUpperCase();
    }*/

    function UpdateReadout() {
        //				var rgb = hsl2rgb(hue, saturation, lightness);
        //				var colorString = normalizedColorToHexString(rgb.r, rgb.g, rgb.b);

        var a = (Math.round(alpha * 100) / 100);
        var l = Math.round(lightness * 100);
        var hslString = Math.round(hue * 360).toString() + ", " + Math.round(saturation * 100).toString() + "%, " + l.toString() + "%";
        var colorString;

        if (a != 1)
            colorString = "hsla(" + hslString + ", " + a.toString() + ")";
        else if (l == 100)
            colorString = "white";
        else if (l == 0)
            colorString = "black";
        else
            colorString = "hsl(" + hslString + ")";

        if (onUpdate)
            onUpdate(colorString);
    }

    function pointerStart(target, pointerId, x, y) {
        var xy = { x: x, y: y };

        if (xy.y > canvasHeight - margin - aArray.height - margin && xy.x < margin + aArray.width + margin) {
            movingAlpha = true;
            ChangeAlphaToCanvasCoords(xy);
        }
        else if (xy.x < margin + hsArray.width + margin) {
            movingHueSaturation = true;
            ChangeHueSaturationToCanvasCoords(xy);
        } else if (xy.y < margin + lArray.height + margin) {
            movingLightness = true;
            ChangeLightnessToCanvasCoords(xy);
        }
        else {
            //  black/white control
            var topLeftX = canvasWidth - margin - lArray.width;
            var topLeftY = canvasHeight - margin - aArray.height;
            var bottomRightX = canvasWidth - margin;
            var bottomRightY = canvasHeight - margin;

            if (xy.x >= topLeftX && xy.x < topLeftX + 12 && xy.y >= topLeftY && xy.y < topLeftY + 12) {
                SetHSLA(hue, 0, 0, 1); // black
            }
            else if (xy.x >= bottomRightX - 12 && xy.x <= bottomRightX && xy.y >= bottomRightY - 12 && xy.y <= bottomRightY) {
                SetHSLA(hue, 0, 1, 1); // white
            }
        }
    }

    function pointerEnd(target, pointerId) {
            movingHueSaturation = false;
            movingLightness = false;
            movingAlpha = false;
    }

    function pointerMove(target, pointerId, x, y) {
        if (movingHueSaturation || movingLightness || movingAlpha) {

            //	tranform mouse event coordinates
            var xy = {x : x, y: y};

            if (movingAlpha) {
                ChangeAlphaToCanvasCoords(xy);
            }
            else if (movingHueSaturation) {
                ChangeHueSaturationToCanvasCoords(xy);
            }
            else {
                ChangeLightnessToCanvasCoords(xy);
            }
        }
    }

    PointerDraw(canvas, pointerStart, pointerMove, pointerEnd);
    
    if (initialColorString)
        SetColorFromString(initialColorString);
    else {
        DrawHueSaturationControl();
        DrawLightnessControl();
        DrawAlphaControl();
        UpdateReadout();
    }

    //  this is the one thing not yet drawn
    DrawBlackWhiteControl();
}
